/*************************************************************************
***     Authentication, authorization, accounting + firewalling package
***     (c) 1998-2003 Anton Vinokurov <anton@netams.com>
***		(c) 2003 NeTAMS Development Team
***		All rights reserved. See 'Copying' file included in distribution
***		For latest version and more info, visit this project web page
***		located at http://www.netams.com
***		 					
*************************************************************************/
/* $Id: ds_netflow.c,v 1.4 2003/04/12 11:27:37 anton Exp $ */

#include "netams.h"

/* Cisco NetFlow FlowCollector
 * Copyright (c) 1997 by Cisco Systems, Inc.
 * All rights reserved.		*/

#define FLOW_VERSION_5 5
#define V5FLOWS_PER_PAK 30

typedef struct Flow5StatHdr {
    ushort version;         
    ushort count;           /* The number of records in PDU. */
    unsigned long SysUptime;       /* Current time in millisecs since router booted */
    unsigned long unix_secs;       /* Current seconds since 0000 UTC 1970 */
    unsigned long unix_nsecs;      /* Residual nanoseconds since 0000 UTC 1970 */
    unsigned long flow_sequence;   /* Seq counter of total flows seen */
    unsigned char engine_type;     /* Type of flow switching engine */
    unsigned char engine_id;       /* ID number of the flow switching engine */
    ushort reserved;
} Flow5StatHdr;


typedef struct IPFlow5Stat {
    in_addr srcaddr;    /* Source IP Address */
    in_addr dstaddr;    /* Destination IP Address */
    in_addr nexthop;    /* Next hop router's IP Address */
    ushort input;          /* Input interface index */
    ushort output;         /* Output interface index */
    
    unsigned long dPkts;           /* Packets sent in Duration */
    unsigned long dOctets;         /* Octets sent in Duration. */
    unsigned long First;           /* SysUptime at start of flow */
    unsigned long Last;            /* and of last packet of flow */
    
    unsigned short srcport;        /* TCP/UDP source port number or equivalent */
    unsigned short dstport;        /* TCP/UDP destination port number or equivalent */
    unsigned char pad;
    unsigned char tcp_flags;      /* Cumulative OR of tcp flags */
    unsigned char prot;           /* IP protocol, e.g., 6=TCP, 17=UDP, ... */
    unsigned char tos;            /* IP Type-of-Service */
    unsigned short src_as;         /* originating AS of source address */
    unsigned short dst_as;         /* originating AS of destination address */
    unsigned char src_mask;       /* source address prefix mask bits */
    unsigned char dst_mask;       /* destination address prefix mask bits */
    ushort reserved;
} IPFlow5Stat;

typedef struct {
    Flow5StatHdr header;
    IPFlow5Stat  records[V5FLOWS_PER_PAK];
} IPStat5Msg;


//////////////////////////////////////////////////////////////////////////
typedef struct vip {
	struct ip iph;
	struct tcphdr tch;
	} vip;
//////////////////////////////////////////////////////////////////////////
unsigned sDS_AcctNetFlow(Service *s, struct ip *ip, struct timeval *tv) {

	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
	struct vip v_ip; // we will treat this flow record as a virtual IP packet

	IPStat5Msg *msg = (IPStat5Msg*)ip;
	IPFlow5Stat *rec;
	Flow5StatHdr *hdr = &msg->header;
	hdr->version = htons(hdr->version);

	char *b1, buf_1[32], *b2, buf_2[32]; // for debugging

	if (hdr->version==FLOW_VERSION_5) {
		hdr->count         = htons(hdr->count);
		hdr->SysUptime     = htonl(hdr->SysUptime);
		hdr->unix_secs     = htonl(hdr->unix_secs);
		hdr->unix_nsecs    = htonl(hdr->unix_nsecs);
		hdr->flow_sequence = htonl(hdr->flow_sequence);

//		aDebug(DEBUG_NF, " cfg seq %lu, fl seq %lu, %d records:\n", cfg->seq_id, hdr->flow_sequence, hdr->count);

		if (!cfg->seq_id) cfg->seq_id=hdr->flow_sequence;
		if (cfg->seq_id!=hdr->flow_sequence) { 
			if (hdr->flow_sequence> V5FLOWS_PER_PAK) aLog(D_WARN, "NF:%d awaited seq %lu, received %lu from %02X (%d flows are lost)\n", cfg->id, cfg->seq_id, hdr->flow_sequence, hdr->engine_id, hdr->flow_sequence-cfg->seq_id);
			cfg->seq_id=hdr->flow_sequence; // restoring back
			}

		cfg->seq_id+=hdr->count;

		aDebug(DEBUG_NF, " NetFlow v%d:%02X packet ds:%d seq %lu/%lu with %d records:\n", hdr->version, hdr->engine_id, cfg->id, hdr->flow_sequence, cfg->seq_id, hdr->count);

		for (int i = 0; i < hdr->count; i++) {
			rec = &msg->records[i];
			
			bzero((void*)&v_ip, sizeof(v_ip));
			
			v_ip.iph.ip_hl=ip->ip_hl;
			v_ip.iph.ip_len=htons(htonl(rec->dOctets));
			v_ip.iph.ip_p=rec->prot;
			memcpy(&v_ip.iph.ip_src, &rec->srcaddr, sizeof(struct in_addr)); 
			memcpy(&v_ip.iph.ip_dst, &rec->dstaddr, sizeof(struct in_addr)); 

    		v_ip.tch.th_sport = rec->srcport;
	    	v_ip.tch.th_dport = rec->dstport;
	
			b1=inet_ntoa(v_ip.iph.ip_src); strncpy(buf_1, b1, strlen(b1)+1);
			b2=inet_ntoa(v_ip.iph.ip_dst); strncpy(buf_2, b2, strlen(b2)+1);
			aDebug(DEBUG_NF, " IP src:%s,dst:%s,l:%ld,p:%d,sp:%u,dp:%u\n", buf_1, buf_2, htonl(rec->dOctets), (v_ip.iph.ip_p), ntohs(v_ip.tch.th_sport), ntohs(v_ip.tch.th_dport));
			if (ntohs(v_ip.iph.ip_len)==htonl(rec->dOctets)) sDS_AcctIp(s, (struct ip*)&v_ip, tv, 1);
			else { 
				sDS_AcctIp(s, (struct ip*)&v_ip, tv, 1, htonl(rec->dOctets));
//				aDebug(DEBUG_NF, " too big packet, %ld should be instead of %d\n", htonl(rec->dOctets), ntohs(v_ip.iph.ip_len));
				}

    		}
	
		}
	
	else aDebug(DEBUG_NF, " NetFlow v%d packet unsupported\n", hdr->version);

	return 0;
	}
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

