/*************************************************************************
***     Authentication, authorization, accounting + firewalling package
***     (c) 1998-2003 Anton Vinokurov <anton@netams.com>
***		(c) 2003 NeTAMS Development Team
***		All rights reserved. See 'Copying' file included in distribution
***		For latest version and more info, visit this project web page
***		located at http://www.netams.com
***		 					
*************************************************************************/
/* $Id: flowprobe.c,v 1.8 2003/04/12 11:27:37 anton Exp $ */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/stat.h>

#ifdef LINUX
#define __FAVOR_BSD 1
        extern "C" {
        #include <pcap.h>                   
	}  
#else 
#include <pcap.h>
#endif

#include <paths.h>
#include <time.h>
#include <signal.h>
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

//---------------------------------------------------------------------------------
/* Cisco NetFlow FlowCollector
 * Copyright (c) 1997 by Cisco Systems, Inc.
 * All rights reserved.		*/

#define FLOW_VERSION_5 5
#define V5FLOWS_PER_PAK 30

typedef struct Flow5StatHdr {
    ushort version;         
    ushort count;           /* The number of records in PDU. */
    unsigned long SysUptime;       /* Current time in millisecs since router booted */
    unsigned long unix_secs;       /* Current seconds since 0000 UTC 1970 */
    unsigned long unix_nsecs;      /* Residual nanoseconds since 0000 UTC 1970 */
    unsigned long flow_sequence;   /* Seq counter of total flows seen */
    unsigned char engine_type;     /* Type of flow switching engine */
    unsigned char engine_id;       /* ID number of the flow switching engine */
    ushort reserved;
} Flow5StatHdr;


typedef struct IPFlow5Stat {
    in_addr srcaddr;    /* Source IP Address */
    in_addr dstaddr;    /* Destination IP Address */
    in_addr nexthop;    /* Next hop router's IP Address */
    ushort input;          /* Input interface index */
    ushort output;         /* Output interface index */
    
    unsigned long dPkts;           /* Packets sent in Duration */
    unsigned long dOctets;         /* Octets sent in Duration. */
    unsigned long First;           /* SysUptime at start of flow */
    unsigned long Last;            /* and of last packet of flow */
    
    unsigned short srcport;        /* TCP/UDP source port number or equivalent */
    unsigned short dstport;        /* TCP/UDP destination port number or equivalent */
    unsigned char pad;
    unsigned char tcp_flags;      /* Cumulative OR of tcp flags */
    unsigned char prot;           /* IP protocol, e.g., 6=TCP, 17=UDP, ... */
    unsigned char tos;            /* IP Type-of-Service */
    unsigned short src_as;         /* originating AS of source address */
    unsigned short dst_as;         /* originating AS of destination address */
    unsigned char src_mask;       /* source address prefix mask bits */
    unsigned char dst_mask;       /* destination address prefix mask bits */
    ushort reserved;
} IPFlow5Stat;

typedef struct {
    Flow5StatHdr header;
    IPFlow5Stat  records[V5FLOWS_PER_PAK];
} IPStat5Msg;

//---------------------------------------------------------------------------------
typedef struct entry;
typedef struct entry {
	u_char ip_tos;
	unsigned long ip_len;
	unsigned long count;
	u_char ip_proto;
	in_addr_t ip_src, ip_dst;
	time_t start;
	time_t last;
	entry *next;
	u_short src_port, dst_port;
	unsigned expired;
	quad_t id;
	} entry;

entry **table;
quad_t last_id=0, expired_id=0;
quad_t sent_packets=0, sent_flows=0;
IPStat5Msg *message=NULL;

#define IPV4_HASH(addr1, addr2, port1, port2, proto) ((((addr1) + (addr2)) + (port1) + (port2)+ (proto)) % ((64*1024)))

//---------------------------------------------------------------------------------
void doalarm(int signal);
void termination(int signal);
void pcap_callback(unsigned char *args, const struct pcap_pkthdr *hdr, const unsigned char *packet);
void Expiresearch();
void Processpacket();
unsigned Hash(struct ip *ip);
unsigned Compare(entry *e, struct ip *ip);
void CheckExpire(entry *e, time_t t);
void PrintUsage(FILE *);
void ToSend(entry *e);
void DoSend();

unsigned char *g_pkt;
unsigned long g_len;
int alarmset=0, offset;
pcap_t *pcap;

char *destination=NULL;
char *dst_host=NULL;
int dst_port=20001;
unsigned t_active=0;
unsigned t_inactive=0;
int udp_socket_fd;
struct sockaddr_in udp_dest;

//---------------------------------------------------------------------------------
int main(int argc, char **argv){
int debug=0, quiet=0;
char *rule=NULL;
char *interface=NULL;
struct bpf_program pf;
bpf_u_int32 net, mask;
char *errbuf;

int op;

while((op=getopt(argc, argv, "hqde:r:i:1:2:")) != EOF){
	switch(op){
		case 'h': PrintUsage(stdout);
				  exit(-1);
				  break;
		case 'd': debug=1;
				  break;
		case 'q': quiet=1;
				  break;
		case 'e': destination=(char *)malloc(strlen(optarg));
				  strcpy(destination, optarg);
				  break;
		case 'r': rule=(char *)malloc(strlen(optarg));
				  strcpy(rule, optarg);
				  break;
		case 'i': interface=(char *)malloc(strlen(optarg));
				  strcpy(interface, optarg);
				  break;
		case '1': t_active=atoi(optarg);
				  break;
		case '2': t_inactive=atoi(optarg);
				  break;
		default: break;
		}
	}

printf("FlowProbe 1.0 (c) 2002 Anton Vinokurov, anton@inorg.chem.msu.ru\n");
printf("This is part of NeTAMS project, http://www.inorg.chem.msu.ru/anton/netams \n");

if (!interface) { PrintUsage(stderr); exit(-1); }  
if (!destination) dst_host="127.0.0.1"; 
else { 
	char *c; c=strchr(destination, ':');
	if (c) { dst_port=atoi(c+1); c[0]='\0'; dst_host=destination; }
	else dst_host=destination;
	}
if (!rule) rule="ip";
if (!t_active) t_active=10*60; // 1 min
if (!t_inactive) t_inactive=1*60; // 10 min

if (!quiet) {
	printf("Exporting to:\t%s:%d\n", dst_host, dst_port);
	printf("Interface:\t%s\n", interface);
	printf("Rule:\t\t%s\n", rule);
	printf("Active timeout:\t%d seconds\n", t_active);
	printf("Inact timeout:\t%d seconds\n", t_inactive);
	}

signal(SIGALRM, (sig_t)doalarm);
signal(SIGINT, (sig_t)termination);

errbuf=(char*)malloc(PCAP_ERRBUF_SIZE);
pcap=pcap_open_live(interface, 64*1024-1, 1, 100, errbuf);

if (!pcap) { 
	if (!quiet) fprintf(stderr, "failed to open pcap interface: %s\n", errbuf);	
	exit(1);	}

if (-1==pcap_lookupnet(interface, &net, &mask, errbuf)) {
	if (!quiet) fprintf(stderr, "failed to lookup net pcap interface: %s\n", errbuf);	
	exit(2);	}


if (-1==pcap_compile(pcap, &pf, rule, 1, net)) { 
	if (!quiet) fprintf(stderr, "failed to compile pcap rule: %s\n", errbuf);	
	exit(3);	}

if (-1==pcap_setfilter(pcap, &pf)) { 
	if (!quiet) fprintf(stderr, "failed to setfilter pcap rule: %s\n", errbuf);	
	exit(4);	}

switch (pcap_datalink(pcap)){ 
	case DLT_NULL:
	offset=0;
	if (!quiet) fprintf(stderr, "Libpcap:\tBSD loopback\n");
	break;

	case DLT_EN10MB:
	offset=ETHER_HDR_LEN;
	if (!quiet) fprintf(stderr, "Libpcap:\tethernet interface\n");
	break;

	case DLT_PPP:
	offset=4; // PPP_HDRLEN
	if (!quiet) fprintf(stderr, "Libpcap:\tPPP interface\n");
	break;

	default:
	offset=0;
	if (!quiet) fprintf(stderr, "Libpcap:\tunknown interface\n");
	break;
	}

bzero(&udp_dest, sizeof(udp_dest));
udp_dest.sin_family=AF_INET;
udp_dest.sin_port=htons(dst_port);
inet_aton(dst_host, &udp_dest.sin_addr);
udp_socket_fd=socket(AF_INET, SOCK_DGRAM, 0);
if (udp_socket_fd==-1) {  
	if (!quiet) fprintf(stderr, "failed to create UDP socket: %s\n", errbuf);	
	exit(5);
	}

table=(entry**)malloc(65536*sizeof(entry*));
for (int i=0; i<65536; i++) table[i]=NULL;

if (!debug) {
	daemon(0, 1); 
	chdir(_PATH_VARTMP);
	}

alarm(3);
while(1) {

	pcap_dispatch(pcap, 1, pcap_callback, NULL);

	if (alarmset) {
		Expiresearch();
		alarm(3);
		alarmset=0;
		}

	if (g_pkt) Processpacket();
	}

} 		
//---------------------------------------------------------------------------------
void pcap_callback(unsigned char *args, const struct pcap_pkthdr *hdr, const unsigned char *packet){
	g_pkt=(unsigned char*)packet;
	g_len=hdr->len-ETHER_HDR_LEN;	
	}

//---------------------------------------------------------------------------------
void termination(int signal){
	printf("\nlast_id=%qu, active=%qu\n", last_id, last_id-expired_id);
	printf("sent %qu flows in %qu packets\n", sent_flows, sent_packets);
	pcap_close(pcap);
	exit(0);
	}

//---------------------------------------------------------------------------------
void doalarm(int signal){
	alarmset=1;
	}

//---------------------------------------------------------------------------------
void Processpacket(){
	struct ip *ip;
	ip=(struct ip*)(g_pkt+offset);
	unsigned hash=Hash(ip);
	entry *e;

	char buf1[32], buf2[32];
	strcpy(buf1, inet_ntoa(ip->ip_src)); strcpy(buf2, inet_ntoa(ip->ip_dst));

	struct tcphdr *th;
	th=(struct tcphdr *)((unsigned char *)ip + ip->ip_hl*4);

	for(e=table[hash]; e!=NULL; e=e->next){
		if (Compare(e, ip)) {
			e->ip_len+=g_len;
			e->count++;
			e->last=time(NULL);
			break;
			}
		}
	 if (e==NULL) { // we must create a new flow record
		e=(entry*)malloc(sizeof(entry));
		e->ip_dst=ip->ip_dst.s_addr;
		e->ip_src=ip->ip_src.s_addr;
		e->ip_len=ntohs(ip->ip_len);  
		e->ip_proto=ip->ip_p;
		e->ip_tos=ip->ip_tos;
		e->start=e->last=time(NULL);
		e->expired=0;
		e->id=last_id;
		e->count=1;
		last_id++;
		e->next=NULL;
		if (e->ip_proto==IPPROTO_TCP || e->ip_proto==IPPROTO_UDP) {
			e->src_port=ntohs(th->th_sport);
			e->dst_port=ntohs(th->th_dport);
			}

		if (!table[hash]) table[hash]=e; 
		else {
			entry *t, *tt; 	
			for (t=table[hash]; t!=NULL; t=t->next) tt=t;
			tt->next=e;
			}
		}

//	printf("[%d(%qu)]-%p %d s:%s:%u d:%s:%u %d\n", hash, e->id, e, ip->ip_p, buf1, ntohs(th->th_sport), buf2, ntohs(th->th_dport), ntohs(ip->ip_len));
	
	g_pkt=0;
	}

//---------------------------------------------------------------------------------
unsigned Compare(entry *e, struct ip *ip) {
	struct tcphdr *th;

	if (e->ip_dst!=ip->ip_dst.s_addr) return 0;
	if (e->ip_src!=ip->ip_src.s_addr) return 0;
	if (e->ip_proto!=ip->ip_p) return 0;
	if (e->ip_tos!=ip->ip_tos) return 0;
	if (e->ip_proto==IPPROTO_TCP || e->ip_proto==IPPROTO_UDP) {
		th=(struct tcphdr *)((unsigned char *)ip + ip->ip_hl*4);
		if (e->src_port!=ntohs(th->th_sport)) return 0;	
		if (e->dst_port!=ntohs(th->th_dport)) return 0;
		}
	return 1;
	}

//---------------------------------------------------------------------------------
void CheckExpire(entry *e, time_t t) {
	if ((unsigned)e->last - (unsigned)e->start > t_active) e->expired=1;
	else if (time(NULL) - (unsigned)e->last > t_inactive) e->expired=1;
	}

//---------------------------------------------------------------------------------
void Expiresearch(){
	int i;
	entry *e, *prev;
	time_t t;

	t=time(NULL);
	printf("\nCheck_expire t:[%qu] a:[%qu] at %ld: ", last_id, last_id-expired_id, t);

	for (i=0; i<65536; i++) {
		JUMP:
		prev=NULL;
		for(e=table[i]; e!=NULL; e=e->next) {
			CheckExpire(e, t);
			if (e->expired) {
				printf("%qu{%d} ", e->id, i); 
  				expired_id++;
				if (e==table[i]) table[i]=e->next;
				else prev->next=e->next;
				ToSend(e);
				free(e);
				goto JUMP;
				}
			prev=e;
			}
		}

	// it's time to check if there is stalled unsent data in message
	if (message && message->header.count) 
		if (message->records[0].Last + 2*t_inactive < (unsigned long)time(NULL))
			DoSend();
	}
//---------------------------------------------------------------------------------
unsigned Hash(struct ip *ip){
	short sp=0, dp=0;
	struct tcphdr *th;

	if (ip->ip_p==IPPROTO_TCP || ip->ip_p==IPPROTO_UDP) {
		th=(struct tcphdr *)((unsigned char *)ip + ip->ip_hl*4);
		sp=ntohs(th->th_sport);
		dp=ntohs(th->th_dport);
		}

	return IPV4_HASH(ip->ip_src.s_addr, ip->ip_dst.s_addr, sp, dp, ip->ip_p);
	}
//---------------------------------------------------------------------------------
void PrintUsage(FILE *f){
	fprintf(f, "Usage: flowprobe {options}\nwhere {options} are:\n \
	-h\t print help screen and exit\n\
	-q\t quiet output\n\
	-d\t debug output\n\
	-e export_to\t IP address to export flows to\n\
	-r rule\t\t libpcap rule to capture packets\n\
	-i interface\t network interface to listen\n\
	-1 active_timeout\t active flow timeout (sec.)\n\
	-2 inactive_timeout\t inactive flow timeout (sec.)\n");

	}
//---------------------------------------------------------------------------------
void ToSend(entry *e){

	if (!message) {
		message=(IPStat5Msg*)malloc(sizeof(IPStat5Msg));
		bzero(message, sizeof(IPStat5Msg));
		message->header.version=ntohs(FLOW_VERSION_5);
		message->header.count=0;
		message->header.engine_id=ntohs(0);
		message->header.engine_type=ntohs(31);
		message->header.flow_sequence=ntohl(sent_flows);
		}

	//struct timeval tv;
	//getmicrouptime(&tv);
	//time_t now=time(NULL);

	message->records[message->header.count].dOctets=ntohl(e->ip_len);
	message->records[message->header.count].dPkts=ntohl(e->count);	
	message->records[message->header.count].input=0;
	message->records[message->header.count].output=0;
	message->records[message->header.count].nexthop.s_addr=INADDR_ANY;

	message->records[message->header.count].tos=e->ip_tos;
	message->records[message->header.count].prot=e->ip_proto;
	message->records[message->header.count].First=e->start;
	message->records[message->header.count].Last=e->last;

	message->records[message->header.count].dst_as=0;
	message->records[message->header.count].dst_mask=0;
	message->records[message->header.count].dstaddr.s_addr=e->ip_dst;
	message->records[message->header.count].dstport=ntohs(e->dst_port);

	message->records[message->header.count].src_as=0;
	message->records[message->header.count].src_mask=0;
	message->records[message->header.count].srcaddr.s_addr=e->ip_src;
	message->records[message->header.count].srcport=ntohs(e->src_port);

	message->header.count++;
	printf("record %d in packet %qu prepared, %ld\n", message->header.count-1, sent_packets+1, e->ip_len);
	if (message->header.count==V5FLOWS_PER_PAK) DoSend();

	}
//---------------------------------------------------------------------------------
void DoSend(){
	sent_packets++;
	sent_flows=sent_flows+message->header.count;
	printf("sending NetFlow v5 packet with %d flows\n", message->header.count);

	message->header.count=ntohs(message->header.count);
	if (sendto(udp_socket_fd, message, sizeof(IPStat5Msg), 0, (sockaddr*)&udp_dest, sizeof(udp_dest)) <0) 
		printf("UDP send failed[%d]: %s", errno, strerror(errno));

	free(message);
	message=NULL;
	}
//---------------------------------------------------------------------------------
